from onegov.core.security import Public
from onegov.winterthur import WinterthurApp, _
from onegov.org.layout import DefaultLayout
from onegov.winterthur.roadwork import RoadworkConnectionError


@WinterthurApp.html(
    model=RoadworkConnectionError,
    permission=Public,
    template='roadwork_connection_error.pt')
def handle_roadwork_connection_error(self, request):

    @request.after
    def set_status_code(response):
        response.status_code = 500

    return {
        'layout': DefaultLayout(self, request),
        'title': _("Connection Error"),
    }



from onegov.winterthur import _

# translations provided by onegov.org, overriden by Winterthur
_("Hide from the public")
_(
    "This site is hidden from the general public. "
    "To show it to everyone, edit this page and disable \"Hide from public\"."
)



from onegov.core.security import Public
from onegov.winterthur import WinterthurApp, _
from onegov.winterthur.daycare import DaycareSubsidyCalculator
from onegov.winterthur.daycare import DaycareSubsidyCalculatorForm
from onegov.winterthur.layout import DaycareSubsidyCalculatorLayout


@WinterthurApp.form(
    model=DaycareSubsidyCalculator,
    form=DaycareSubsidyCalculatorForm,
    permission=Public,
    template='daycare.pt')
def view_daycare_subsidy_calculator(self, request, form):
    calculation = None

    if form.submitted(request):
        calculation = self.calculate(
            daycare=form.selected_daycare,
            services=form.services.services,
            income=form.income.data,
            wealth=form.wealth.data,
            rebate=form.rebate.data
        )

    return {
        'title': _("Daycare Subsidy Calculator"),
        'layout': DaycareSubsidyCalculatorLayout(self, request),
        'form': form,
        'calculation': calculation,
        'button_text': _("Calculate"),
        'settings': self.settings,
        'eligible': (
            calculation and calculation.city_share_per_month != '0.00'
        )
    }



from onegov.core.orm import Base
from onegov.core.orm.abstract import associated
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.types import UUID, UTCDateTime
from onegov.file import File
from onegov.org.models import HiddenFromPublicExtension
from sedate import to_timezone
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Numeric
from sqlalchemy import Text
from sqlalchemy.orm import relationship
from uuid import uuid4


class MissionReportFile(File):
    __mapper_args__ = {'polymorphic_identity': 'mission-report-file'}


class MissionReport(Base, ContentMixin, HiddenFromPublicExtension):

    __tablename__ = 'mission_reports'

    #: the public id of the mission_report
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the date of the report
    date = Column(UTCDateTime, nullable=False)

    #: how long the mission lasted, in hours
    duration = Column(Numeric(precision=6, scale=2), nullable=False)

    #: the nature of the mission
    nature = Column(Text, nullable=False)

    #: the location of the mission
    location = Column(Text, nullable=False)

    #: actually active personnel
    personnel = Column(Integer, nullable=False)

    #: backup personnel
    backup = Column(Integer, nullable=False)

    #: the Zivilschutz was involved
    civil_defence = Column(Boolean, nullable=False, default=False)

    #: the vehicle use of the mission report
    used_vehicles = relationship(
        'MissionReportVehicleUse',
        cascade="all, delete-orphan")

    #: pictures of the mission
    pictures = associated(MissionReportFile, 'pictures', 'one-to-many')

    @property
    def title(self):
        return self.nature

    @property
    def readable_duration(self):
        return str(self.duration).rstrip('.0') + 'h'

    @property
    def local_date(self):
        return to_timezone(self.date, 'Europe/Zurich')


class MissionReportVehicle(Base, ContentMixin, HiddenFromPublicExtension):

    __tablename__ = 'mission_report_vehicles'

    #: the public id of the vehicle
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the short id of the vehicle
    name = Column(Text, nullable=False)

    #: the longer name of the vehicle
    description = Column(Text, nullable=False)

    #: symbol of the vehicle
    symbol = associated(MissionReportFile, 'symbol', 'one-to-one')

    #: a website describing the vehicle
    website = Column(Text, nullable=True)

    uses = relationship('MissionReportVehicleUse')

    @property
    def title(self):
        return f'{self.name} - {self.description}'

    @property
    def readable_website(self):
        if self.website:
            return self.website.replace('https://', '').replace('http://', '')


class MissionReportVehicleUse(Base):
    """ Many to many association between vehicles and reports. """

    __tablename__ = 'mission_report_vehicle_usees'

    mission_report_id = Column(
        UUID,
        ForeignKey('mission_reports.id'),
        primary_key=True)

    mission_report = relationship('MissionReport')

    vehicle_id = Column(
        UUID,
        ForeignKey('mission_report_vehicles.id'),
        primary_key=True)

    vehicle = relationship('MissionReportVehicle')

    # vehicles may be used multiple times in a single mission_report
    count = Column(
        Integer,
        nullable=False,
        default=1)



from onegov.file.utils import IMAGE_MIME_TYPES_AND_SVG
from onegov.form import Form
from onegov.form.fields import TimezoneDateTimeField
from onegov.form.fields import UploadFileWithORMSupport
from onegov.form.validators import FileSizeLimit
from onegov.form.validators import WhitelistedMimeType
from onegov.winterthur import _
from onegov.winterthur.models import MissionReportFile
from wtforms.fields import BooleanField
from wtforms.fields import StringField
from wtforms.fields import TextAreaField
from wtforms.fields.html5 import DecimalField
from wtforms.fields.html5 import IntegerField
from wtforms.fields.html5 import URLField
from wtforms.validators import InputRequired
from wtforms.validators import NumberRange
from wtforms.validators import Optional


class MissionReportForm(Form):

    date = TimezoneDateTimeField(
        _("Date"),
        timezone='Europe/Zurich',
        validators=[InputRequired()])

    duration = DecimalField(
        _("Mission duration (h)"),
        validators=[InputRequired(), NumberRange(0, 10000)])

    nature = TextAreaField(
        _("Mission nature"),
        render_kw={'rows': 4},
        validators=[InputRequired()])

    location = StringField(
        _("Location"),
        validators=[InputRequired()])

    personnel = IntegerField(
        _("Mission personnel"),
        validators=[InputRequired(), NumberRange(0, 10000)])

    backup = IntegerField(
        _("Backup personnel"),
        validators=[InputRequired(), NumberRange(0, 10000)])

    civil_defence = BooleanField(
        _("Civil Defence involvement"))


class MissionReportVehicleForm(Form):

    name = StringField(
        _("Name"),
        validators=[InputRequired()])

    description = StringField(
        _("Description"),
        validators=[InputRequired()])

    symbol = UploadFileWithORMSupport(
        _("Symbol"),
        file_class=MissionReportFile,
        validators=[
            Optional(),
            WhitelistedMimeType(IMAGE_MIME_TYPES_AND_SVG),
            FileSizeLimit(1 * 1024 * 1024)
        ])

    website = URLField(
        _("Website"))



from onegov.core.elements import Link
from onegov.core.security import Public, Private
from onegov.form import FieldDependency, WTFormsClassBuilder, move_fields
from onegov.org.views.files import view_get_image_collection
from onegov.winterthur import WinterthurApp, _
from onegov.winterthur.collections import MissionReportCollection
from onegov.winterthur.collections import MissionReportFileCollection
from onegov.winterthur.collections import MissionReportVehicleCollection
from onegov.winterthur.forms import MissionReportForm
from onegov.winterthur.forms import MissionReportVehicleForm
from onegov.winterthur.layout import MissionReportLayout
from onegov.winterthur.models import MissionReport
from onegov.winterthur.models import MissionReportVehicle
from onegov.winterthur.models import MissionReportVehicleUse
from uuid import UUID
from wtforms.fields import BooleanField
from wtforms.fields.html5 import IntegerField


def mission_report_form(model, request):
    if isinstance(model, MissionReportCollection):
        report = MissionReport()
    else:
        report = model

    form_class = report.with_content_extensions(MissionReportForm, request)

    class MissionReportVehicleUseForm(form_class):

        def populate_obj(self, obj):
            super().populate_obj(obj)

            for used in obj.used_vehicles:
                request.session.delete(used)

            request.session.flush()

            fids = (
                fid for fid in self._fields

                if fid.startswith('vehicles_')
                and not fid.endswith('_count')
                and self.data[fid]
            )

            for fid in fids:

                obj.used_vehicles.append(
                    MissionReportVehicleUse(
                        vehicle_id=UUID(fid.replace('vehicles_', '')),
                        count=self.data[f'{fid}_count']))

        def process_obj(self, obj):
            super().process_obj(obj)

            for used in obj.used_vehicles:
                field_id = f'vehicles_{used.vehicle_id.hex}'

                getattr(self, field_id).data = True
                getattr(self, f'{field_id}_count').data = used.count

    builder = WTFormsClassBuilder(MissionReportVehicleUseForm)
    builder.set_current_fieldset(_("Vehicles"))

    vehicles = MissionReportVehicleCollection(request.session).query()
    vehicles = {v.id: v for v in vehicles if not v.is_hidden_from_public}

    # include hidden vehicles that were picked before being hidden
    for used in report.used_vehicles:
        if used.vehicle_id not in vehicles:
            vehicles[used.vehicle_id] = used.vehicle

    vehicle_field_id = None

    for vehicle in sorted(vehicles.values(), key=lambda v: v.name):
        field_id = f'vehicles_{vehicle.id.hex}'
        vehicle_field_id = f'{field_id}_count'

        builder.add_field(
            field_class=BooleanField,
            field_id=field_id,
            label=vehicle.title,
            required=False,
            id=vehicle.id
        )

        builder.add_field(
            field_class=IntegerField,
            field_id=vehicle_field_id,
            label=request.translate(_("Count")),
            required=True,
            dependency=FieldDependency(field_id, 'y'),
            default=1
        )

    form_class = builder.form_class

    if vehicle_field_id:
        form_class = move_fields(
            form_class, ('is_hidden_from_public', ), vehicle_field_id)

    return form_class


def mission_report_vehicle_form(model, request):
    if isinstance(model, MissionReportVehicleCollection):
        report = MissionReportVehicle()
    else:
        report = model

    return report.with_content_extensions(MissionReportVehicleForm, request)


@WinterthurApp.html(
    model=MissionReportCollection,
    permission=Public,
    template='mission_reports.pt')
def view_mission_reports(self, request):
    return {
        'layout': MissionReportLayout(self, request),
        'title': _("Mission Reports"),
        'reports': self.batch,
        'count': self.mission_count(),
        'year': self.year,
    }


@WinterthurApp.html(
    model=MissionReport,
    permission=Public,
    template='mission_report.pt')
def view_mission(self, request):
    return {
        'title': self.title,
        'layout': MissionReportLayout(
            self, request,
            Link(self.title, '#')
        ),
        'model': self
    }


@WinterthurApp.html(
    model=MissionReportVehicleCollection,
    permission=Private,
    template='mission_report_vehicles.pt')
def view_mission_report_vehicles(self, request):

    return {
        'layout': MissionReportLayout(
            self, request,
            Link(_("Vehicles"), request.link(self))
        ),
        'title': _("Vehicles"),
        'vehicles': tuple(self.query()),
    }


@WinterthurApp.html(model=MissionReportFileCollection, template='images.pt',
                    permission=Private)
def view_mission_report_files(self, request):
    data = view_get_image_collection(self, request)
    data['layout'] = MissionReportLayout(
        self, request,
        Link(self.report.title, request.link(self.report)),
        Link(_("Images"), '#', editbar=False)
    )

    return data


@WinterthurApp.form(
    model=MissionReportCollection,
    permission=Private,
    form=mission_report_form,
    name='new',
    template='form.pt')
def handle_new_mission_report(self, request, form):

    if form.submitted(request):
        mission = self.add(**{
            k: v for k, v in form.data.items()
            if k != 'csrf_token'
            and not k.startswith('vehicles_')
        })

        form.populate_obj(mission)

        request.success(_("Successfully added a mission report"))
        return request.redirect(request.link(mission))

    return {
        'title': _("Mission Reports"),
        'form': form,
        'layout': MissionReportLayout(
            self, request,
            Link(_("New"), '#', editbar=False)
        ),
    }


@WinterthurApp.form(
    model=MissionReport,
    permission=Private,
    form=mission_report_form,
    name='edit',
    template='form.pt')
def handle_edit_mission_report(self, request, form):

    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))
        return request.redirect(request.link(self))

    elif not request.POST:
        form.process(obj=self)

    return {
        'title': _("Mission Reports"),
        'form': form,
        'layout': MissionReportLayout(
            self, request,
            Link(self.title, request.link(self)),
            Link(_("Edit"), '#', editbar=False),
        )
    }


@WinterthurApp.view(
    model=MissionReport,
    permission=Private,
    request_method='DELETE')
def delete_mission_report(self, request):
    request.assert_valid_csrf_token()
    request.session.delete(self)
    request.success(_("Successfully deleted mission report"))


@WinterthurApp.form(
    model=MissionReportVehicleCollection,
    permission=Private,
    form=mission_report_vehicle_form,
    name='new',
    template='form.pt')
def handle_new_vehicle(self, request, form):

    if form.submitted(request):
        vehicle = self.add(
            **{k: v for k, v in form.data.items() if k not in (
                'csrf_token', 'symbol'
            )})

        # required for the symbol image
        form.populate_obj(vehicle)

        request.success(_("Successfully added a vehicle"))

        return request.redirect(
            request.class_link(MissionReportVehicleCollection))

    return {
        'title': _("Vehicle"),
        'form': form,
        'layout': MissionReportLayout(
            self, request,
            Link(_("Vehicles"), request.link(self)),
            Link(_("New"), '#', editbar=False)
        ),
    }


@WinterthurApp.form(
    model=MissionReportVehicle,
    permission=Private,
    form=mission_report_vehicle_form,
    name='edit',
    template='form.pt')
def handle_edit_vehicle(self, request, form):

    if form.submitted(request):
        form.populate_obj(self)

        request.success(_("Your changes were saved"))

        return request.redirect(
            request.class_link(MissionReportVehicleCollection))

    elif not request.POST:
        form.process(obj=self)

    return {
        'title': self.title,
        'form': form,
        'layout': MissionReportLayout(
            self, request,
            Link(_("Vehicles"), request.class_link(
                MissionReportVehicleCollection)),
            Link(self.title, '#')
        )
    }


@WinterthurApp.view(
    model=MissionReportVehicle,
    permission=Private,
    request_method='DELETE')
def delete_mission_report_vehicle(self, request):
    request.assert_valid_csrf_token()
    request.session.delete(self)
    request.success(_("Successfully deleted vehicle"))



import isodate
import pycurl
import sedate

from cached_property import cached_property
from datetime import datetime, timedelta
from io import BytesIO
from onegov.core.custom import json
from pathlib import Path
from purl import URL


class RoadworkError(Exception):
    pass


class RoadworkConnectionError(RoadworkError):
    pass


class RoadworkConfig(object):
    """ Looks at ~/.pdb.secret and /etc/pdb.secret (in this order), to extract
    the configuration used for the RoadworkClient class.

    The configuration is as follows::

        HOSTNAME: pdb.example.org
        ENDPOINT: 127.0.0.1:6004
        USERNAME: username
        PASSWORD: password

    * The HOSTNAME is the address of the PDB service.
    * The ENDPOINT is the optional address of the tcp-proxy used.
    * The USERNAME is the NTLM password.
    * The PASSWORD is the NTLM password.

    """

    def __init__(self, hostname, endpoint, username, password):
        self.hostname = hostname
        self.endpoint = endpoint
        self.username = username
        self.password = password

    @classmethod
    def lookup_paths(self):
        yield Path('~/.pdb.secret').expanduser()
        yield Path('/etc/pdb.secret')

    @classmethod
    def lookup(cls):
        for path in cls.lookup_paths():
            if path.exists():
                return cls(**cls.parse(path))

        paths = ', '.join(str(p) for p in cls.lookup_paths())
        raise RoadworkError(
            f"No pdb configuration found in {paths}")

    @classmethod
    def parse(cls, path):
        result = {
            'hostname': None,
            'endpoint': None,
            'username': None,
            'password': None,
        }

        with path.open('r') as file:
            for line in file:
                line = line.strip()

                if not line:
                    continue

                if ':' not in line:
                    continue

                if line.startswith('#'):
                    continue

                k, v = line.split(':', maxsplit=1)
                k = k.strip().lower()
                v = v.strip()

                if k in result:
                    result[k] = v

        return result


class RoadworkClient(object):
    """ A proxy to Winterthur's internal roadworks service. Uses redis as
    a caching mechanism to ensure performance and reliability.

    Since the roadworks service can only be reached inside Winterthur's
    network, we rely on a proxy connection during development/testing.

    To not expose any information unwittingly to the public, the description
    of how to connect to that proxy is kept at docs.seantis.ch.

    """

    def __init__(self, cache, hostname, username, password, endpoint=None):
        self.cache = cache
        self.hostname = hostname
        self.username = username
        self.password = password
        self.endpoint = endpoint or hostname

    @cached_property
    def curl(self):
        curl = pycurl.Curl()
        curl.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_NTLM)
        curl.setopt(pycurl.USERPWD, f"{self.username}:{self.password}")
        curl.setopt(pycurl.HTTPHEADER, [f'HOST: {self.hostname}'])

        return curl

    def url(self, path):
        return f'http://{self.endpoint}/{path}'

    def get(self, path, lifetime=5 * 60, downtime=60 * 60):
        """ Requests the given path, returning the resulting json if
        successful.

        A cache is used in two stages:

        * At the lifetime stage, the cache is returned unconditionally.
        * At the end of the lifetime, the cache is refreshed if possible.
        * At the end of the downtime stage the cache forcefully refreshed.

        During its lifetime the object is basically up to 5 minutes out of
        date. But since the backend may not be available when that time
        expires we operate with a downtime that is higher (1 hour).

        This means that a downtime in the backend will not result in evicted
        caches, even if the lifetime is up. Once the downtime limit is up we
        do however evict the cache forcefully, raising an error if we cannot
        connect to the backend.

        """
        path = path.lstrip('/')

        cached = self.cache.get(path)

        def refresh():
            try:
                status, body = self.get_uncached(path)
            except pycurl.error:
                raise RoadworkConnectionError(
                    f"Could not connect to {self.hostname}")

            if status == 200:
                self.cache.set(path, {
                    'created': datetime.utcnow(),
                    'status': status,
                    'body': body
                })

                return body

            raise RoadworkError(f"{path} returned {status}")

        # no cache yet, return result and cache it
        if not cached:
            return refresh()

        now = datetime.utcnow()
        lifetime_horizon = cached['created'] + timedelta(seconds=lifetime)
        downtime_horizon = cached['created'] + timedelta(seconds=downtime)

        # within cache lifetime, return cached value
        if now <= lifetime_horizon:
            return cached['body']

        # outside cache lifetime, but still in downtime horizon, try to
        # refresh the value but ignore errors
        if lifetime_horizon < now < downtime_horizon:
            try:
                return refresh()
            except RoadworkConnectionError:
                return cached['body']

        # outside the downtime lifetime, force refresh and raise errors
        return refresh()

    def get_uncached(self, path):
        body = BytesIO()

        self.curl.setopt(pycurl.URL, self.url(path))
        self.curl.setopt(pycurl.WRITEFUNCTION, body.write)
        self.curl.perform()

        status = self.curl.getinfo(pycurl.RESPONSE_CODE)
        body = body.getvalue().decode('utf-8')

        if status == 200:
            body = json.loads(body)

        return status, body

    def is_cacheable(self, response):
        return response[0] == 200


class RoadworkCollection(object):

    def __init__(self, client, letter=None, query=None):
        self.client = client
        self.query = None
        self.letter = None

        if query:
            self.query = query.lower()

        elif letter:
            self.letter = letter.lower()

    @property
    def letters(self):
        letters = set()

        for roadwork in self.by_letter(None).roadwork:
            for letter in roadwork.letters:
                letters.add(letter)

        letters = list(letters)
        letters.sort()

        return letters

    def by_filter(self, filter):

        # note: addGisLink doesn't work here
        url = URL('odata/Baustellen')\
            .query_param('addGisLink', 'False')\
            .query_param('$filter', filter)

        records = self.client.get(url.as_string()).get('value', ())
        records = (r for r in records if r['Internet'])

        work = [Roadwork(r) for r in records]
        work.sort(key=lambda r: r.title)

        return work

    @property
    def roadwork(self):
        date = datetime.today()

        roadwork = self.by_filter(filter=' and '.join((
            f'DauerVon le {date.strftime("%Y-%m-%d")}',
            f'DauerBis ge {date.strftime("%Y-%m-%d")}',
        )))

        # The backend supports searches/filters, but the used dataset is
        # so small that it makes little sense to use that feature, since it
        # would lead to a lot more cache-misses on our end.
        #
        # Instead we simply loop through the results and filter them out.
        if self.query:
            roadwork = [
                r for r in roadwork if self.query in r.title.lower()
            ]

        elif self.letter:
            roadwork = [
                r for r in roadwork if self.letter in r.letters
            ]

        return roadwork

    def by_id(self, id):
        url = URL(f'odata/Baustellen({int(id)})')\
            .query_param('addGisLink', 'True')

        work = tuple(
            Roadwork(r) for r in self.client.get(
                url.as_string()).get('value', ()))

        if work:
            return work[0]

        # secondary lookup is against the subsections.. this probably calls
        # for an index eventually
        for r in self.roadwork:
            for section in r.sections:
                if section.id == id:
                    return section

    def by_letter(self, letter):
        return self.__class__(self.client, letter=letter, query=None)


class Roadwork(object):

    def __init__(self, data):
        self.data = data

        self.convertors = {
            'DauerVon': lambda v: v and isodate.parse_datetime(v),
            'DauerBis': lambda v: v and isodate.parse_datetime(v),
        }

    @property
    def id(self):
        return self['Id']

    @property
    def letters(self):
        for key in ('ProjektBezeichnung', 'ProjektBereich'):
            if self[key]:
                letter = self[key][0].lower()

                if letter in ('abcdefghijklmnopqrstuvwxyz'):
                    yield letter

    @property
    def title(self):
        parts = (self[key] for key in ('ProjektBezeichnung', 'ProjektBereich'))
        parts = (p.strip() for p in parts if p)
        parts = (p for p in parts)

        return ' '.join(parts)

    @property
    def sections(self):
        now = sedate.utcnow()

        sections = (
            self.__class__({
                'Id': r['TeilbaustelleId'],
                'Teilbaustellen': [],
                **r
            }) for r in self['Teilbaustellen']
        )

        sections = (s for s in sections if s['DauerVon'])
        sections = (s for s in sections if s['DauerVon'] <= now)
        sections = (s for s in sections if now <= (s['DauerBis'] or now))

        return list(sections)

    def __getitem__(self, key):
        value = self.data[key]

        if key in self.convertors:
            return self.convertors[key](value)

        return value

    def __contains__(self, key):
        return key in self.data



from onegov.winterthur.collections.address import AddressCollection
from onegov.winterthur.collections.address import AddressSubsetCollection

from onegov.winterthur.collections.mission_report \
    import MissionReportCollection

from onegov.winterthur.collections.mission_report \
    import MissionReportFileCollection

from onegov.winterthur.collections.mission_report \
    import MissionReportVehicleCollection

__all__ = (
    'AddressCollection',
    'AddressSubsetCollection',
    'MissionReportCollection',
    'MissionReportFileCollection',
    'MissionReportVehicleCollection'
)



import sedate

from datetime import datetime, date
from onegov.core.collection import GenericCollection, Pagination
from onegov.org.models.file import ImageFileCollection
from onegov.winterthur.models import MissionReport
from onegov.winterthur.models import MissionReportFile
from onegov.winterthur.models import MissionReportVehicle
from sqlalchemy import and_, or_, desc


class MissionReportFileCollection(ImageFileCollection):

    def __init__(self, session, report):
        super().__init__(session)
        self.type = 'mission-report-file'
        self.report = report

    @property
    def id(self):
        return self.report.id

    @property
    def association(self):
        return MissionReportFile.registered_links['linked_mission_reports']

    def add(self, *args, **kwargs):
        file = super().add(*args, **kwargs)
        self.report.pictures.append(file)

        return file

    def query(self):
        query = super().query()
        table = self.association.table

        query = query.filter(MissionReportFile.id.in_(
            self.session.query(table)
                .with_entities(table.c.missionreportfile_id)
                .filter(table.c.mission_reports_id == self.report.id)
                .subquery()
        ))

        return query


class MissionReportCollection(GenericCollection, Pagination):

    def __init__(self, session, page=0, include_hidden=False, year=None):
        self.session = session
        self.page = page
        self.include_hidden = include_hidden
        self.year = year or date.today().year

    @property
    def model_class(self):
        return MissionReport

    def __eq__(self, other):
        return self.page == other.page

    def query(self):
        query = super().query()

        if not self.include_hidden:
            query = query.filter(or_(
                MissionReport.meta['is_hidden_from_public'] == False,
                MissionReport.meta['is_hidden_from_public'] == None
            ))

        query = self.filter_by_year(query)

        return query.order_by(desc(MissionReport.date))

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(self.session, page=index, year=self.year)

    def filter_by_year(self, query):
        timezone = 'Europe/Zurich'

        start = sedate.replace_timezone(datetime(self.year, 1, 1), timezone)
        end = sedate.replace_timezone(datetime(self.year + 1, 1, 1), timezone)

        return query.filter(and_(
            start <= MissionReport.date, MissionReport.date < end
        ))

    def mission_count(self):
        """ The mission count, including hidden missions. """

        return self.filter_by_year(super().query()).count()


class MissionReportVehicleCollection(GenericCollection):

    @property
    def model_class(self):
        return MissionReportVehicle

    def query(self):
        return super().query().order_by(MissionReportVehicle.name)



import click
import json
import sedate
import shutil

from datetime import datetime
from onegov.core.cli import command_group
from onegov.core.crypto import random_token
from onegov.core.csv import CSVFile
from onegov.file.utils import as_fileintent
from onegov.winterthur.models import MissionReport
from onegov.winterthur.models import MissionReportFile
from onegov.winterthur.models import MissionReportVehicle
from onegov.winterthur.models import MissionReportVehicleUse
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile


cli = command_group()


@cli.command(
    name='import-mission-reports', context_settings={'singular': True})
@click.option('--vehicles-file', type=click.Path(exists=True), required=True)
@click.option('--missions-file', type=click.Path(exists=True), required=False)
@click.option('--no-confirm', is_flag=True, default=False)
def import_mission_reports(vehicles_file, missions_file, no_confirm):
    """ Imports the existing mission reports. """

    if not no_confirm:
        click.confirm(
            'This will delete all existing missions and vehicles, continue?',
            abort=True
        )

    with open(vehicles_file, 'rb') as f:
        vehicles = tuple(CSVFile(f, expected_headers=(
            'uid', 'name', 'description', 'hidden'
        )))

    with open(missions_file, 'rb') as f:
        missions = tuple(CSVFile(f, expected_headers=(
            'date',
            'alert_time',
            'duration',
            'type',
            'location',
            'personnel',
            'personnel_hq',
            'vehicles',
            'public'
        )))

    def extract_date(mission):
        d = datetime.utcfromtimestamp(int(mission.date))
        d = sedate.replace_timezone(d, 'Europe/Zurich')

        time = mission.alert_time.replace('.', ':')
        time = time.replace(';', ':')
        time = time.replace('::', ':')

        if ':' in time:
            h, m = (int(v.strip('<>-')) for v in time.split(':'))
        else:
            h = int(time)
            m = 0

        d = d.replace(hour=h)
        d = d.replace(minute=m)

        return d

    def extract_duration(mission):
        d = mission.duration
        d = d.replace(',', '.')

        if ':' in d:
            hours, minutes = (int(v) for v in d.split(':'))
            return hours + minutes / 60

        return float(d)

    def extract_personnel(value):

        def digits():
            for v in value:
                if v.isdigit():
                    yield v
                else:
                    break

        return int(''.join(digits()))

    def is_hidden(mission):
        if mission.public == '0':
            return True

        if mission.hidden == '1':
            return True

        if mission.hide == '1':
            return True

        if "nicht freischalten" in mission.type.lower():
            return True

        return False

    def handle_import(request, app):

        for mission in request.session.query(MissionReport):
            request.session.delete(mission)

        for vehicle in request.session.query(MissionReportVehicle):
            request.session.delete(vehicle)

        created_vehicles = {}

        for vehicle in vehicles:
            created = MissionReportVehicle(
                name=vehicle.name,
                description=vehicle.description,
            )

            if vehicle.hidden == '1':
                created.is_hidden_from_public = True

            created_vehicles[vehicle.uid] = created
            request.session.add(created)

        request.session.flush()

        for mission in missions:
            created = MissionReport(
                date=extract_date(mission),
                duration=extract_duration(mission),
                nature=mission.type,
                location=mission.location,
                personnel=extract_personnel(mission.personnel),
                backup=extract_personnel(mission.personnel_hq),
                meta={'is_hidden_from_public': is_hidden(mission)}
            )

            vehicle_uids = mission.vehicles.split(',')
            vehicle_uids = (v.strip() for v in vehicle_uids)
            vehicle_uids = (v for v in vehicle_uids if v)

            for uid in vehicle_uids:
                created.used_vehicles.append(MissionReportVehicleUse(
                    vehicle=created_vehicles[uid]
                ))

            request.session.add(created)

        request.session.flush()

    return handle_import


@cli.command(
    name='export-mission-vehicles', context_settings={'singular': True})
@click.option('--export-file', type=click.Path(exists=False), required=True)
def export_mission_vehicles(export_file):
    """ Exports the mission vehicles (with symbols, but without usage) into
    a ZIP file for consumption with the 'import-mission-vehicles' command.

    """

    def handle_export(request, app):
        temp = TemporaryDirectory()
        path = Path(temp.name)
        data = {}

        for vehicle in request.session.query(MissionReportVehicle):
            record = data[str(vehicle.id)] = {
                'name': vehicle.name,
                'description': vehicle.description,
                'website': vehicle.website,
                'filename': None
            }

            if vehicle.symbol:
                symbols = (path / 'symbols')
                symbols.mkdir(exist_ok=True)

                record['filename'] = vehicle.symbol.name

                with (symbols / str(vehicle.id)).open('wb') as f:
                    shutil.copyfileobj(vehicle.symbol.reference.file, f)

        with (path / 'data.json').open('w') as f:
            json.dump(data, f, sort_keys=True, indent=4)

        shutil.make_archive(export_file.rstrip('.zip'), 'zip', path)

    return handle_export


@cli.command(
    name='import-mission-vehicles', context_settings={'singular': True})
@click.option('--import-file', type=click.Path(exists=True), required=True)
@click.option('--replace', is_flag=True, default=False)
@click.option('--match', is_flag=True, default=False)
def import_mission_vehicles(import_file, replace, match):
    """ Imports the mission vehicles created by the export-mission-vehicles
    command.

    """

    def handle_import(request, app):
        temp = TemporaryDirectory()
        path = Path(temp.name)
        data = {}

        zip = ZipFile(import_file)
        zip.extractall(path)

        with (path / 'data.json').open('r') as f:
            data = json.load(f)

        existing = {
            str(v.id): v for v in request.session.query(MissionReportVehicle)
        }

        for id in data:

            if id in existing and replace:
                request.session.delete(existing[id])
                request.session.flush()

            if id in existing and not replace:
                continue

            vehicle = None

            if match:
                vehicle = request.session.query(MissionReportVehicle)\
                    .filter_by(name=data[id]['name'])\
                    .first()

            if not vehicle:
                vehicle = MissionReportVehicle(id=id)

            vehicle.name = data[id]['name']
            vehicle.description = data[id]['description']
            vehicle.website = data[id]['website']

            symbol_path = path / 'symbols' / str(id)

            if symbol_path.exists():
                filename = data[id]['filename']

                with symbol_path.open('rb') as f:
                    vehicle.symbol = MissionReportFile(
                        id=random_token(),
                        name=filename,
                        reference=as_fileintent(
                            content=f,
                            filename=filename
                        )
                    )

            request.session.add(vehicle)

    return handle_import
